// Form1.cpp

#include "stdafx.h"

#include "Form1.h"
#include "WndMessageHolder.h"

#include <CommCtrl.h>

using namespace BringUpTest;

// Message handler for primary instance activation

void Form1::WndProc(Message %msg)
{
  WndMessageHolder ^wmh = WndMessageHolder::GetInstance();
  if (msg.Msg == wmh->WM_BRINGUP) BringMeUp();
  else if (msg.Msg == wmh->WM_MESSAGEECHO_ADVERTISE) MessageEchoAdvertiseHandler(msg.WParam);
  else if (msg.Msg == wmh->WM_MESSAGEECHO_ADVERTISEMENTRESPONSE) MessageEchoAdvertisementResponseHandler(msg.WParam);
  __super::WndProc(msg);
}

// Handles WM_BRINGUP message

void Form1::BringMeUp()
{
  switch (m_ism) {
    case InstanceSwitchMethod::Classic:
      // "Classic" version according to TheGreatCthulhu's suggestion - unfortunately doesn't
      // really work completely... (see CG forum thread)

      if (WindowState == FormWindowState::Minimized)
        WindowState = FormWindowState::Normal;

      Activate();

      break;

    case InstanceSwitchMethod::Alternative:
      // Alternative version - works but unfortunately looks hackish...

      if (WindowState != FormWindowState::Minimized)
        WindowState = FormWindowState::Minimized;
      WindowState = FormWindowState::Normal;
      // Activate();
      break;

    case InstanceSwitchMethod::FakeHotkey:
      // Actually the simplest of them all:
      // Just pass a WM_SYSCOMMAND with SC_HOTKEY to myself by calling the WndProc()

      WndProc(Message::Create(Handle, WM_SYSCOMMAND, (IntPtr)SC_HOTKEY, Handle));
      break;

    default:
      Trace::Fail("Message Echo method and Deposited Handle method don't use the WM_BRINGUP message");
      break;
  }
}

// Handles WM_MESSAGEECHO_ADVERTISE message - gets passed the non-primary instance main window handle

void Form1::MessageEchoAdvertiseHandler(IntPtr ipHwndNonPrimary)
{
  bool bResult = WndMessageHolder::PostMessage_(ipHwndNonPrimary,
    WndMessageHolder::GetInstance()->WM_MESSAGEECHO_ADVERTISEMENTRESPONSE, Handle, (IntPtr)0);
  Trace::Assert(bResult, String::Format("Posting WM_MESSAGEECHO_ADVERTISEMENTRESPONSE message failed (code {0})",
    WndMessageHolder::GetLastError_()));
}

// Handles WM_MESSAGEECHO_ADVERTISEMENTRESPONSE message - gets passed the primary instance main window handle

void Form1::MessageEchoAdvertisementResponseHandler(IntPtr ipHwndPrimary)
{
  bool bResult = WndMessageHolder::BringWindowToTop_(ipHwndPrimary);
  Trace::Assert(bResult, String::Format("Calling BringWindowToTop() on primary instance main window handle failed (code {0})",
    WndMessageHolder::GetLastError_()));
  tmrMessageEchoTimeout->Stop();
  Close();
}

// Event handlers to switch between instance switching modes

System::Void Form1::rbClassic_CheckedChanged(System::Object^  sender, System::EventArgs^  e)
{
  if (rbClassic->Checked) m_ism = InstanceSwitchMethod::Classic;
}

System::Void Form1::rbAlt_CheckedChanged(System::Object^  sender, System::EventArgs^  e)
{
  if (rbAlt->Checked) m_ism = InstanceSwitchMethod::Alternative;
}

System::Void Form1::rbMessageEcho_CheckedChanged(System::Object^  sender, System::EventArgs^  e)
{
  if (rbMessageEcho->Checked) {
    if (!m_evtMessageEchoSignal) {
      m_evtMessageEchoSignal = gcnew EventWaitHandle(true, EventResetMode::ManualReset, m_strMessageEchoSignalEvent);
    } else m_evtMessageEchoSignal->Set();
    m_ism = InstanceSwitchMethod::MessageEcho;
  } else m_evtMessageEchoSignal->Reset();
}

System::Void Form1::rbDepositedHandle_CheckedChanged(System::Object^  sender, System::EventArgs^  e)
{
  if (rbDepositedHandle->Checked) {
    Application::UserAppDataRegistry->SetValue("PrimaryWindowHandle", int(Handle));
    m_bRegistryEntryCreated = true;
    m_ism = InstanceSwitchMethod::DepositedHandle;
  } else RemoveRegistryEntry();
}

System::Void Form1::rbFakeHotkey_CheckedChanged(System::Object^  sender, System::EventArgs^  e)
{
  if (rbFakeHotkey->Checked) m_ism = InstanceSwitchMethod::FakeHotkey;
}

// Tear down our entire registry branch in order to leave zero registry footprint

void Form1::RemoveRegistryEntry()
{
  if (m_bRegistryEntryCreated) {
    RegistryKey ^rkCompany = Registry::CurrentUser->OpenSubKey("Software\\" + Application::CompanyName, true);
    rkCompany->DeleteSubKeyTree(Application::ProductName);
    if (rkCompany->SubKeyCount + rkCompany->ValueCount == 0) {  // Delete company key if it's empty now
      rkCompany->Close();
      Registry::CurrentUser->OpenSubKey("Software", true)->DeleteSubKey(Application::CompanyName);
    }
    m_bRegistryEntryCreated = false;
  }
}

System::Void Form1::Form1_Load(System::Object^  sender, System::EventArgs^  e)
{
  String ^strInstanceMutex = "{BE3A098C-D815-43e7-B8E2-5EC55CBC5306}";
#ifdef _DEBUG
  strInstanceMutex += "_debug";
#endif

  bool bIMTheOne;
  m_mtxInstance = gcnew Mutex(true, strInstanceMutex, bIMTheOne);
  if (!bIMTheOne) {
    WndMessageHolder ^wmh = WndMessageHolder::GetInstance();

    // Don't use Application::UserAppDataRegistry here to avoid creating the key if it doesn't exist already:

    Object ^objRegValue = Registry::GetValue(String::Format("HKEY_CURRENT_USER\\Software\\{0}\\{1}\\{2}",
      Application::CompanyName, Application::ProductName, Application::ProductVersion), "PrimaryWindowHandle", nullptr);
    if (objRegValue) {

      // The registry value exists, so use the Deposited Handle method. Calling BringWindowToTop()
      // on the stored window handle from here doesn't report failure but has no effect either -
      // for unknown reasons. Posting a message to ourselves works but has no visible advantage
      // over the Message Echo method.

      bool bResult = WndMessageHolder::PostMessage_(this, wmh->WM_MESSAGEECHO_ADVERTISEMENTRESPONSE,
        IntPtr(safe_cast<int>(objRegValue)), (IntPtr)0);
      Trace::Assert(bResult, String::Format("Posting WM_MESSAGEECHO_ADVERTISEMENTRESPONSE message to self failed (code {0})",
        WndMessageHolder::GetLastError_()));

    } else {
      m_evtMessageEchoSignal = gcnew EventWaitHandle(false, EventResetMode::ManualReset, m_strMessageEchoSignalEvent);

      if (m_evtMessageEchoSignal->WaitOne(0)) {

        // The event isn't actually part of the Message Echo mechanism. It's just used in this demo
        // to signal the non-primary instance to use the Message Echo method.

        // Advertise our own main window handle:

        Trace::WriteLine("BringUpTest already running, sending WM_MESSAGEECHO_ADVERTISE message...");
        bool bResult = WndMessageHolder::PostMessage_((IntPtr)HWND_BROADCAST, wmh->WM_MESSAGEECHO_ADVERTISE, Handle, (IntPtr)0);
        Trace::Assert(bResult, String::Format("Broadcasting WM_MESSAGEECHO_ADVERTISE message failed (code {0})",
          WndMessageHolder::GetLastError_()));

        // The primary instance is now supposed to respond with a WM_MESSAGEECHO_ADVERTISEMENTRESPONSE message
        // carrying its own main window handle

        // Make my own main window invisible and start timeout:

        Visible = false;
        tmrMessageEchoTimeout->Start();

      } else {

        Trace::WriteLine("BringUpTest already running, sending WM_BRINGUP message...");
        bool bResult = WndMessageHolder::PostMessage_((IntPtr)HWND_BROADCAST, wmh->WM_BRINGUP, (IntPtr)0, (IntPtr)0);
        Trace::Assert(bResult, String::Format("Broadcasting WM_BRINGUP message failed (code {0})",
          WndMessageHolder::GetLastError_()));
        Close();

      }
    }
  } else {
    // Get myself a hotkey

    IntPtr lres =
      WndMessageHolder::SendMessage_(this, WM_SETHOTKEY, (IntPtr)((HOTKEYF_ALT | HOTKEYF_CONTROL) << 16 | 'B'), (IntPtr)0);
    Trace::WriteLine(String::Format("Sending WM_SETHOTKEY to self returned {0}", lres));
  }
}

System::Void Form1::tmrMessageEchoTimeout_Tick(System::Object^  sender, System::EventArgs^  e)
{
  // This handler actually should never get called

  Trace::Fail("BringUpTest: Message Echo mechanism timed out");
}
